#!/usr/bin/env python

"""
OpenID provider login resources which redirect clients back to the application
("relying party").

Copyright (C) 2004, 2005, 2006, 2007 Paul Boddie <paul@boddie.org.uk>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

import WebStack.Generic
from WebStack.Helpers.Auth import Authenticator, check_openid_signature, make_openid_signature
import datetime
import time
import random

class OpenIDLoginResource:

    "A resource providing a login screen."

    encoding = "utf-8"
    openid_ns = "http://specs.openid.net/auth/2.0"

    def __init__(self, app_url, authenticator, associations=None, use_redirect=1, urlencoding=None, encoding=None):

        """
        Initialise the resource with the application URL 'app_url' and an
        'authenticator'.

        The optional 'associations' is a mapping from association handles to
        secret keys.

        If the optional 'use_redirect' flag is set to 0, a confirmation screen
        is given instead of redirecting the user back to the original
        application.

        The optional 'urlencoding' parameter allows a special encoding to be
        used in producing the redirection path.

        The optional 'encoding' parameter allows a special encoding to be used
        in producing the login pages.

        To change the pages employed by this resource, either redefine the
        'login_page' and 'success_page' attributes in instances of this class or
        a subclass, or override the 'show_login' and 'show_success' methods.
        """

        self.app_url = app_url
        self.authenticator = authenticator
        self.associations = associations or {}
        self.use_redirect = use_redirect
        self.urlencoding = urlencoding
        self.encoding = encoding or self.encoding

    def respond(self, trans):

        "Respond using the transaction 'trans'."

        # Check for a submitted login form.

        fields_body = trans.get_fields_from_body(self.encoding)

        if fields_body.has_key("login"):

            # Check a combination of local identifier and username together with
            # the password.

            claimed_id = fields_body.get("claimed_id", [""])[0]
            local_id = fields_body.get("local_id", [""])[0]
            username = fields_body.get("username", [""])[0]
            password = fields_body.get("password", [""])[0]
            app = fields_body.get("app", [""])[0]

            # NOTE: Permit flexibility in the credentials.

            if self.authenticator.authenticate(trans, (local_id, username), password):
                self._redirect(trans, claimed_id, local_id, username, app)
                # The above method does not return.

        # Check for an OpenID signature verification request.

        elif fields_body.get("openid.mode", [None])[0] == "check_authentication":

            # Obtain the secret key from recorded associations.

            handle = fields_body.get("openid.assoc_handle", [None])[0]
            if handle is not None and self.associations.has_key(handle):
                valid = check_openid_signature(fields_body, self.associations[handle])
                del self.associations[handle]
            else:
                valid = 0

            # Produce a response for this request.

            self.show_verification(trans, valid)
            # The above method does not return.

        # NOTE: Permit association requests here.

        # Otherwise, show the login form.

        fields_path = trans.get_fields_from_path(self.urlencoding)
        app = fields_path.get("openid.return_to", [""])[0]
        claimed_id = fields_path.get("openid.claimed_id", [""])[0]
        local_id = fields_path.get("openid.identity", [""])[0]

        self.show_login(trans, app, claimed_id, local_id)

    def _redirect(self, trans, claimed_id, local_id, username, app):

        """
        Redirect the client using 'trans', 'claimed_id', 'local_id', 'username'
        and the given 'app' details.
        """

        app_url = self.app_url + trans.get_path_without_query(self.urlencoding)

        # Make an association that can be used in signature verification.
        # NOTE: Probably need to consider the secret key a bit more.

        handle = username + str(time.time())
        secret_key = str(random.randint(0, 999999999))
        self.associations[handle] = secret_key

        # Make a timestamp.

        now = datetime.datetime.utcnow()
        timestamp = now.strftime("%Y-%m-%dT%H:%M:%SZ") + str(now.microsecond)

        # Make a signature.

        signed_names = ["op_endpoint", "return_to", "response_nonce", "assoc_handle", "claimed_id", "identity"]
        fields = {
            "openid.op_endpoint" : [app_url],
            "openid.return_to" : [app],
            "openid.response_nonce" : [timestamp],
            "openid.assoc_handle" : [handle],
            "openid.claimed_id" : [claimed_id],
            "openid.identity" : [local_id]
            }

        signature = make_openid_signature(signed_names, fields, secret_key)

        # Build an URL for returning to the application.

        url = "%s?openid.ns=%s&openid.mode=%s&openid.signed=%s&openid.sig=%s" % (
            app,
            trans.encode_path(self.openid_ns, self.urlencoding),
            trans.encode_path("id_res", self.urlencoding),
            trans.encode_path(",".join(signed_names), self.urlencoding),
            trans.encode_path(signature, self.urlencoding)
            )

        for name, value in fields.items():
            url += "&%s=%s" % (name, trans.encode_path(value[0], self.urlencoding))

        # Show the success page anyway.
        # NOTE: Offer a POST-based form for redirection.

        self.show_success(trans, url)
        if self.use_redirect:
            trans.redirect(url)
        else:
            raise WebStack.Generic.EndOfResponse

    def show_verification(self, trans, status):

        """
        Writes a signature verification response using the transaction 'trans'
        and the 'status' of the verification.
        """

        trans.set_content_type(WebStack.Generic.ContentType("text/plain"))
        out = trans.get_response_stream()

        # NOTE: Need to use invalidate_handle, too.

        if status:
            status_str = "true"
        else:
            status_str = "false"
        out.write("ns:%s\nis_valid:%s\n" % (self.openid_ns, status_str))
        raise WebStack.Generic.EndOfResponse

    def show_login(self, trans, app, claimed_id, local_id):

        """
        Writes a login screen using the transaction 'trans', including details
        of the 'app' which the client was attempting to access, along with the
        'claimed_id' and 'local_id'.
        """

        trans.set_content_type(WebStack.Generic.ContentType("text/html", self.encoding))
        out = trans.get_response_stream()
        out.write(self.login_page % (app, claimed_id, local_id))

    def show_success(self, trans, app):

        """
        Writes a success screen using the transaction 'trans', including details
        of the 'app' which the client was attempting to access.
        """

        trans.set_content_type(WebStack.Generic.ContentType("text/html", self.encoding))
        out = trans.get_response_stream()
        out.write(self.success_page % (app, app))

    login_page = """
<html>
  <head>
    <title>Login</title>
  </head>
  <body>
    <h1>Login</h1>
    <form method="POST">
      <p>Username: <input name="username" type="text" size="12"/></p>
      <p>Password: <input name="password" type="password" size="12"/></p>
      <p><input name="login" type="submit" value="Login"/></p>
      <input name="app" type="hidden" value="%s"/>
      <input name="claimed_id" type="hidden" value="%s"/>
      <input name="local_id" type="hidden" value="%s"/>
    </form>
  </body>
</html>
"""

    success_page = """
<html>
  <head>
    <title>Login Example</title>
  </head>
  <body>
    <h1>Login Successful</h1>
    <p>Please proceed to the application: <a href="%s">%s</a></p>
  </body>
</html>
"""

# vim: tabstop=4 expandtab shiftwidth=4
